import {desiredHeight, desiredWidth, getScreen} from "./engine/internal/screen.ts";
import {BG_INSET, BG_OUTER, FG_TEXT} from "./colors.ts";
import {checkGrid, ConceptualCell, maps, mapSzX, mapSzY} from "./maps.ts";
import {D, I} from "./engine/public.ts";
import {Grid, IGame, Point, Size} from "./engine/datatypes.ts";
import {sprRaccoonWalking, sprStatPickup} from "./sprites.ts";
import {DrawPile} from "./drawpile.ts";

class MenuCamera {
  // measured in whole screens
  position: Point;
  target: Point;

  constructor({position, target}: {position: Point, target: Point}) {
    this.position = position;
    this.target = target;
  }

  update() {
    let adjust = (x0: number, x1: number) => {
      if (Math.abs(x1 - x0) < 0.01) { return x1; }
      return (x0 * 8 + x1 * 2) / 10;
    }
    this.position = new Point(
      adjust(this.position.x, this.target.x),
      adjust(this.position.y, this.target.y),
    );
  }
}

type GameState = "Gameplay" | "Thralls";

function getScreenLocation(state: GameState): Point {
  if (state === "Gameplay") {
    return new Point(0, 0);
  }
  if (state === "Thralls") {
    return new Point(0, 1);
  }

  throw `invalid state: ${state}`
}

export class Game implements IGame {
  camera: MenuCamera;
  state: GameState;
  huntMode: HuntMode;
  frame: number;

  constructor() {
    this.camera = new MenuCamera({
      position: new Point(0, 0),
      target: new Point(0, 0),
    });
    this.state = "Gameplay";

    this.huntMode = HuntMode.generate({depth: 1});
    this.frame = 0;
  }

  update() {
    this.frame += 1;
    if (I.isKeyPressed("w")) {
      this.state = "Gameplay"
    }
    if (I.isKeyPressed("s")) {
      this.state = "Thralls"
    }

    this.camera.target = getScreenLocation(this.state);
    D.camera = new Point(
      D.size.w * this.camera.position.x,
      D.size.h * this.camera.position.y,
    )
    this.camera.update();

    // state-specific updates
    this.updateGameplay();
  }

  draw() {
    // draw screen background
    let oldCamera = D.camera;
    D.camera = new Point(0, 0);
    D.fillRect(new Point(0, 0), D.size, BG_OUTER);
    D.camera = oldCamera;

    this.drawGameplay();

    // we draw all states at once and pan between them
    // mainFont.drawText({ctx: ctx, text: "You have been given a gift.", x: 0, y: 0})
    let mouse = I.mousePosition?.offset(D.camera);
    if (mouse != null) {
      D.invertRect(mouse.offset(new Point(-1, -1)), new Size(3, 3))
    }
  }

  getPaneRegionForGameState(gameState: GameState) {
    let screen = getScreen();
    let {w, h} = screen.size;
    let overallScreenLocation = getScreenLocation(gameState);

    let bigPaneX = overallScreenLocation.x * w;
    let bigPaneY = overallScreenLocation.y * h;
    let bigPaneW = w;
    let bigPaneH = h;

    let smallPaneW = desiredWidth;
    let smallPaneH = desiredHeight;
    let smallPaneX = Math.floor(bigPaneX + (bigPaneW - smallPaneW) / 2)
    let smallPaneY = Math.floor(bigPaneY + (bigPaneH - smallPaneH) / 2)

    return {
      big: {
        position: new Point(bigPaneX, bigPaneY),
        size: new Size(bigPaneW, bigPaneH),
      },
      small: {
        position: new Point(smallPaneX, smallPaneY),
        size: new Size(smallPaneW, smallPaneH),
      }
    }

  }

  updateGameplay() {

  }

  drawGameplay() {
    let region = this.getPaneRegionForGameState("Gameplay")
    // TODO: Draw
    let oldCamera = D.camera;
    D.camera = D.camera.offset(region.small.position.negate());

    let drawpile = new DrawPile();

    let globalOffset =
      new Point(this.huntMode.player.x * 32, this.huntMode.player.y * 32).offset(
        new Point(-192, -128)
      )


    for (let y = 0; y < mapSzY; y += 1) {
      for (let x = 0; x < mapSzX; x += 1) {
        let cellOffset = new Point(x * 32, y * 32).offset(globalOffset.negate());
        let cellData = this.huntMode.cells.get(new Point(x, y))

        this.#drawMapCell(drawpile, cellOffset, new Point(x, y), cellData);
      }
    }
    this.#drawPlayer(drawpile, globalOffset);

    drawpile.execute();

    D.drawText("hello", new Point(0, 0), FG_TEXT);

    D.camera = oldCamera;
  }

  #drawMapCell(
    drawpile: DrawPile,
    cellOffset: Point,
    mapPosition: Point,
    cellData: MapCell,
  ) {
    const OFFSET_FLOOR = -256;
    const OFFSET_AIR = 0;
    const depth = cellOffset.y;
    const onFloor = OFFSET_FLOOR + depth;
    const inAir = OFFSET_AIR + depth;

    if (cellData.content.type == "block") {
      return;
    }
    /*
    if (!cellData.revealed) {
      return;
    }
     */

    // draw inset zone
    drawpile.add(onFloor, () =>
      D.fillRect(cellOffset.offset(new Size(-16, -26)), new Size(32, 32), BG_INSET)
    );

    /*
    if (!cellData.revealed) {
      // TODO: draw some kind of question mark
      D.drawText("?", cellOffset.offset(new Point(12, 8)), FG_TEXT);
      return
    }
     */
    if (cellData.content.type == "statPickup") {
      let content = cellData.content;
      let extraXOffset = 0; // Math.cos(this.frame / 80 + mapPosition.x + mapPosition.y) * 1;
      let extraYOffset = 0; // Math.sin(this.frame / 50 + mapPosition.x * 2+ mapPosition.y * 0.75) * 6 - 3;
      drawpile.add(inAir, () => {
        D.drawSprite(
          sprStatPickup,
          cellOffset.offset(new Point(extraXOffset, extraYOffset)),
          ALL_STATS.indexOf(content.stat)
        )
      });
    }
  }

  #drawPlayer(drawpile: DrawPile, globalOffset: Point) {
    let cellOffset = new Point(this.huntMode.player.x * 32, this.huntMode.player.y * 32).offset(globalOffset.negate())
    drawpile.add(this.huntMode.player.y, () => {
      D.drawSprite(
        sprRaccoonWalking,
        cellOffset
      )
    });
  }
}

type Stat = "AGI" | "INT" | "CHA" | "PSI";
const ALL_STATS: Array<Stat> = ["AGI", "INT", "CHA", "PSI"];
type MapCellContent =
  {type: "statPickup", stat: Stat} |
  {type: "stairs"} |
  {type: "empty"} |
  {type: "block"}

type MapCell = {
  content: MapCellContent,
  isValidSpawn: boolean,
  revealed: boolean
}

class HuntMode {
  depth: number
  cells: Grid<MapCell>
  player: Point

  constructor({depth, cells, player}: {depth: number, cells: Grid<MapCell>, player: Point }) {
    this.depth = depth;
    this.cells = cells;
    this.player = player;

    checkGrid(this.cells);
  }

  static generate({depth}: {depth: number}) {
    let mapNames: Array<string> = Object.keys(maps);
    let mapName = mapNames[Math.floor(Math.random() * mapNames.length)];
    let map = maps[mapName];

    let cells = map.map((ccell, _xy) => {
      return this.#generateCell(ccell);
    })

    let validSpawns = [];
    for (let x = 0; x < cells.size.w; x++) {
      for (let y = 0; y < cells.size.h; y++) {
        let position = new Point(x, y);
        if (cells.get(position).isValidSpawn)  {
          validSpawns.push(position);
        }
      }
    }
    let player = choose(validSpawns);
    cells.get(player).content = {type: "empty"};

    if (Math.random() < 0.75) {
      while (true) {
        let x = Math.floor(Math.random() * mapSzX);
        let y = Math.floor(Math.random() * mapSzY);
        let xy = new Point(x, y);

        let item = cells.get(new Point(x, y));
        if (player.equals(xy)) {
          continue;
        }
        if (item.content.type == "block") {
          continue;
        }
        item.content = {type: "stairs"}
        break;
      }
    }

    return new HuntMode({depth, cells, player})
  }

  static #generateCell(conceptual: ConceptualCell): MapCell {
    switch (conceptual) {
      case "X":
        return { content: {type: "block"}, revealed: true, isValidSpawn: false};
      case " ":
        return { content: HuntMode.#generateContent(), revealed: false, isValidSpawn: false };
      case ".":
        return { content: HuntMode.#generateContent(), revealed: false, isValidSpawn: true };
    }
  }

  static #generateContent(): MapCellContent {
    // stat pickup
    let gsp = (): MapCellContent => {
      return {type: "statPickup", stat: choose(ALL_STATS)}
    };
    // TODO: Other objects?
    return choose([
      gsp, gsp, gsp, gsp
    ])();
  }
}

function choose<T>(array: Array<T>): T  {
  if (array.length == 0) {
    throw `array cannot have length 0 for choose`
  }
  return array[Math.floor(Math.random() * array.length)]
}

export let game = new Game();